#!/usr/bin/env python

import optparse
import warnings
import logging
import rrddatastore
import forecast
import re
import datetime
import numpy as np

warnings.filterwarnings('ignore')

try:
    import json
except ImportError:
    import simplejson as json

logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s %(levelname)-8s %(message)s',
                    datefmt='%a, %d %b %Y %H:%M:%S',
                    filename='/usr/local/nagiosxi/var/components/capacityplanning.log',
                    filemode='a')


def parse_period(period):
    try:
        period = int(period)
    except TypeError:
        period = 1
    except ValueError:
        WEEKS_IN_YEAR = 52
        WEEKS_IN_MONTH = 4

        period = str(period)
        result = re.search(r'.*([0-9]+)(.*)', period)
        if result:
            prefix = int(result.group(1))
            unit = result.group(2).strip().lower()
            if 'year' in unit:
                period = prefix * WEEKS_IN_YEAR
            elif 'month' in unit:
                period = prefix * WEEKS_IN_MONTH
            else:
                period = prefix
        else:
            period = 1
    return period


def parse_args():
    """Parse the input args of the script. """

    parser = optparse.OptionParser()

    hostgroup = optparse.OptionGroup(
        parser,
        "Data Selection Options",
        "Use this to specify the host/service/track."
    )
    hostgroup.add_option(
        "-H",
        "--host",
        help="Host name to run report with."
    )
    hostgroup.add_option(
        "-S",
        "--service",
        help="Service name to run report with."
    )
    hostgroup.add_option("-T", "--track", help="Track name to run report with.")
    parser.add_option_group(hostgroup)

    manipulation = optparse.OptionGroup(
        parser,
        "Extrapolation Options",
        "Use these to specify how the data will be extrapolated."
    )
    manipulation.add_option(
        "-M",
        "--method",
        help="Holt-Winters, Linear Fit, Quadratic Fit or Cubic Fit"
    )
    manipulation.add_option(
        "-P",
        "--period",
        help="Period of data to use as a basis for future data."
    )
    manipulation.add_option(
        "-E",
        "--end",
        help="End data timestamp for the date to extrapolate out from.",
        default="now"
    )
    manipulation.add_option(
        "-s",
        "--steps",
        type="int",
        help="Number of next steps to extrapolate out, this will extrapolate out n periods."
    )
    manipulation.add_option(
        "-d",
        "--dates",
        help="Comma-delimited list of unix timestamps that are within the range of extrapolation."
    )
    parser.add_option_group(manipulation)

    options, args = parser.parse_args()

    options.period = parse_period(options.period)

    if not options.host:
        parser.error('Host is a required argument.')
    if not options.service:
        parser.error('Service is a required argument.')
    if not options.track:
        parser.error('Track is a required argument.')

    if not options.method in ['Holt-Winters', 'Linear Fit', 'Quadratic Fit', 'Cubic Fit']:
        options.method = 'Holt-Winters'
    if options.steps is None:
        options.steps = 1
    if options.period is None:
        options.period = 1

    if not options.dates is None:
        #[datetime.datetime.fromtimestamp(0) + datetime.timedelta(seconds=int(x)) for x in options.dates.split(',')]
        options.dates = [int(x) for x in options.dates.split(',')]
        try:
            pass
        except Exception, e:
            parser.error('Dates did not make sense. They must be unix timestamps.')
    else:
        options.dates = []

    return options


def main():
    o = parse_args()

    data = rrddatastore.RRDDatastore(o.host, o.service, o.track)
    data.parse_rrd_file(o.period * 2, o.end)
    data.parse_xml_file()
    dataset = data.dataset

    dataset.set_period(o.period)

    if o.method == 'Holt-Winters':
        extrapolated, f, residues = forecast.holt_winters(dataset, o.steps)
    elif o.method == 'Linear Fit':
        extrapolated, f, residues = forecast.least_squares(dataset, o.steps)
    elif o.method == 'Quadratic Fit':
        extrapolated, f, residues = forecast.quadratic_fit(dataset, o.steps)
    elif o.method == 'Cubic Fit':
        extrapolated, f, residues = forecast.cubic_fit(dataset, o.steps)

    dmean = dataset.get_mean()
    dmax  = dataset.get_max()
    emean = extrapolated.get_mean()
    emax  = extrapolated.get_max()

    real = dataset.get_highcharts_dataset()
    predicted = extrapolated.get_highcharts_dataset()

    highcharts_data = [real, predicted]
    dates = [{'date': datetime.datetime.fromtimestamp(x).strftime('%c'),
    #dates = [{'date': x,
              'value': extrapolated.get_value_on_date(x)
             } for x in o.dates]


    info = {}

    info.update(forecast.make_projections(extrapolated, f, residues))

    info.update({
        'emax': emax,
        'emean': emean,
        'dmax': dmax,
        'dmean': dmean,
        'unit': dataset.unit,
        'integrity': dataset.get_integrity(),
        'highcharts': highcharts_data,
        'dates': dates,
    })

    print json.dumps(info, allow_nan=False)


if __name__ == '__main__':
    try:
        main()
    except Exception, e:
        logging.exception(e)
        info = {
            'emax': None,
            'emean': None,
            'dmax': None,
            'dmean': None,
            'integrity': 0,
            'highcharts': [],
            'error': str(e),
        }
        print json.dumps(info)
